@@ -67,26 +67,33 @@ class RequestHandler(object):
         r = self.session.post(url, data=data, params=params)
         return self._response_parser(r, expect_json=False)
 
-    def send(self, request):
+    def send(self, request, expect_json=True, ignore_content=False):
         """
         Send a formatted API request
 
         :param request: a formatted request object
         :type request: :class:`.Request`
+        :param bool expect_json: if True, raise :class:`.InvalidAPIAccess` if
+            response is not in JSON format
+        :param bool ignore_content: whether to ignore setting content of the
+            Response object
         """
         r = self.session.request(method=request.method,
                                  url=request.url,
                                  params=request.params,
                                  data=request.data,
-                                 files=request.files)
-        return self._response_parser(r)
+                                 files=request.files,
+                                 headers=request.headers)
+        return self._response_parser(r, expect_json, ignore_content)
 
-    def _response_parser(self, r, expect_json=True):
+    def _response_parser(self, r, expect_json=True, ignore_content=False):
         """
         :param :class:`requests.Response` r: a response object of the Requests
             library
-        :param bool expect_json: if True, raise :class`.InvalidAPIAccess` if
+        :param bool expect_json: if True, raise :class:`.InvalidAPIAccess` if
             response is not in JSON format
+        :param bool ignore_content: whether to ignore setting content of the
+            Response object
         """
         if r.ok:
             try:
@@ -96,9 +103,14 @@ class RequestHandler(object):
                 # No JSON-encoded data returned
                 if expect_json:
                     logger = logging.getLogger(conf.LOGGING_API_LOGGER)
-                    logger.debug(r.content)
+                    logger.debug(r.text)
                     raise InvalidAPIAccess('Invalid API access.')
-                return Response(False, r.content)
+                # Raw response
+                if ignore_content:
+                    res = Response(True, None)
+                else:
+                    res = Response(True, r.text)
+                return res
         else:
             r.raise_for_status()
 
@@ -176,6 +188,7 @@ class API(object):
     web_api_url = 'http://web.api.115.com/files'
     aps_natsort_url = 'http://aps.115.com/natsort/files.php'
     proapi_url = 'http://proapi.115.com/app/chrome/down'
+    referer_url = 'http://115.com'
 
     def __init__(self, persistent=False,
                  cookies_filename=None, cookies_type='LWPCookieJar'):
@@ -596,6 +609,15 @@ class API(object):
             'ac': 'space',
             '_': get_timestamp(13)
         }
+        _sign = os.environ.get('U115_BROWSER_SIGN')
+        if _sign is not None:
+            _time = os.environ.get('U115_BROWSER_TIME')
+            if _time is None:
+                msg = 'U115_BROWSER_TIME is required given U115_BROWSER_SIGN.'
+                raise APIError(msg)
+            params['sign'] = _sign
+            params['time'] = _time
+            params['uid'] = self.user_id
         req = Request(url=url, params=params)
         r = self.http.send(req)
         if r.state:
@@ -734,6 +756,17 @@ class API(object):
         else:
             raise RequestFailure('Failed to delete the task.')
 
+    def _req_file_userfile(self):
+
+        url = 'http://115.com/'
+        params = {
+            'ct': 'file',
+            'ac': 'userfile',
+            'is_wl_tpl': 1,
+        }
+        req = Request(method='GET', url=url, params=params)
+        self.http.send(req, expect_json=False, ignore_content=True)
+
     def _req_aps_natsort_files(self, cid, offset, limit, o='file_name',
                                asc=1, aid=1, show_dir=1, code=None, scid=None,
                                snap=0, natsort=1, source=None, type=0,
@@ -868,13 +901,19 @@ class API(object):
             raise RequestFailure('No directory found.')
 
     def _req_files_download_url(self, pickcode, proapi=False):
+        if '_115_curtime' not in self.cookies:
+            self._req_file_userfile()
         if not proapi:
             url = self.web_api_url + '/download'
             params = {'pickcode': pickcode, '_': get_timestamp(13)}
         else:
             url = self.proapi_url
             params = {'pickcode': pickcode, 'method': 'get_file_url'}
-        req = Request(method='GET', url=url, params=params)
+        headers = {
+            'Referer': self.referer_url,
+        }
+        req = Request(method='GET', url=url, params=params,
+                      headers=headers)
         res = self.http.send(req)
         if res.state:
             if not proapi:
@@ -1392,13 +1431,16 @@ class Directory(BaseFile):
         kwargs['natsort'] = 1 if natsort is True else 0
         kwargs['o'] = order
 
-        # When the downloads directory exists, its parent directory's count
-        # does not include the downloads directory. This behavior is similar
-        # to the parent's parent (root).
+        # When the downloads directory exists along with its parent directory,
+        # the receiver directory, its parent's count (receiver directory's
+        # count) does not include the downloads directory. This behavior is
+        # similar to its parent's parent (root), the count of which does not
+        # include the receiver directory.
         # The following code fixed this behavior so that a directory's
         # count correctly reflects the actual number of entries in it
         # The side-effect that this code may ensure that downloads directory
-        # exists.
+        # exists, causing the system to create the receiver directory and
+        # downloads directory, if they do not exist.
 
         if self.is_root or self == self.api.receiver_directory:
             self._count += 1
@@ -1471,7 +1513,7 @@ class Task(Base):
 
     def __init__(self, api, add_time, file_id, info_hash, last_update,
                  left_time, move, name, peers, percent_done, rate_download,
-                 size, status, cid, pid):
+                 size, status, cid, pid, url, *args, **kwargs):
         self.api = api
         self.cid = cid
         self.name = name
@@ -1487,6 +1529,7 @@ class Task(Base):
         self.size = size
         self.size_human = humanize.naturalsize(size, binary=True)
         self.status = status
+        self.url = url
         self._directory = None
         self._deleted = False
         self._count = -1
@@ -1494,9 +1537,14 @@ class Task(Base):
     @property
     def is_directory(self):
         """
-        :return: whether this task is associated with a directory
+        :return: whether this task is associated with a directory.
         :rtype: bool
         """
+        if self.cid is None:
+            msg = 'Cannot determine whether this task is a directory.'
+            if not self.is_transferred:
+                msg += ' This task has not been transferred.'
+            raise TaskError(msg)
         return self.api.downloads_directory.cid != self.cid
 
     @property
@@ -1693,7 +1741,7 @@ def _instantiate_task(api, kwargs):
     """Create a Task object from raw kwargs"""
     file_id = kwargs['file_id']
     kwargs['file_id'] = file_id if str(file_id).strip() else None
-    kwargs['cid'] = kwargs['file_id']
+    kwargs['cid'] = kwargs['file_id'] or None
     kwargs['rate_download'] = kwargs['rateDownload']
     kwargs['percent_done'] = kwargs['percentDone']
     kwargs['add_time'] = get_utcdatetime(kwargs['add_time'])
@@ -1705,6 +1753,11 @@ def _instantiate_task(api, kwargs):
         kwargs['pid'] = None
     del kwargs['rateDownload']
     del kwargs['percentDone']
+    if 'url' in kwargs:
+        if not kwargs['url']:
+            kwargs['url'] = None
+    else:
+        kwargs['url'] = None
     task = Task(api, **kwargs)
     if is_transferred:
         task._parent = api.downloads_directory
